use std::fs::{File, OpenOptions};
use std::io::{Read, Write};
use std::time::Duration;

use async_std::task;
use serde::{Deserialize, Serialize};
use sha2::{Digest, Sha256};

use crate::utils::{get_env, host_port, HostPort};

mod utils;

#[derive(Debug, Serialize, Deserialize)]
struct Job {
    job_name: String,
    rule_type: String,
    metrics_path: String,
    targets: Vec<String>,
}

#[derive(Debug, Serialize, Deserialize)]
struct Jobs {
    jobs: Vec<Job>,
    hash: String,
}

impl Jobs {
    fn new() -> Self {
        Jobs {
            jobs: vec![],
            hash: String::new(),
        }
    }

    fn hash(s: String) -> String {
        let mut hasher = Sha256::new();
        hasher.update(s);
        let result = hasher.finalize();
        let hex_digest = result
            .iter()
            .map(|byte| format!("{:02x}", byte))
            .collect::<String>();
        return hex_digest;
    }

    fn load(&mut self) {
        let mut rule_file = File::open("rule.yaml").unwrap();
        let mut rule_file_content = String::new();
        rule_file.read_to_string(&mut rule_file_content).unwrap();
        let new_hash = Self::hash(rule_file_content.clone());
        if self.hash == new_hash {
            println!("配置没有变化");
            return;
        }
        println!("配置变化, 开始刷新配置");
        // 写入配置文件中
        let j = serde_yaml::from_str(&*rule_file_content);
        match j {
            Err(why) => panic!("unmarshal rule.yaml failed: {:?}", why),
            Ok(v) => {self.jobs = v }
        }
        self.hash = new_hash;
        // 定义配置文件
        let mut prometheus = Prometheus::new();
        let job_iter = self.jobs.iter();
        for job in job_iter {
            let mut sc = ScrapeConfig::new(job.job_name.clone(), job.metrics_path.clone());
            if job.rule_type == "static" {
                let static_config = StaticConfig {
                    targets: job.targets.clone(),
                };
                sc.static_configs.push(static_config);
            } else {
                let target_iter = job.targets.iter();
                let mut port: u16 = 0;
                let mut domains: Vec<String> = Vec::new();
                for ti in target_iter {
                    let hp = host_port(ti.clone());
                    match hp {
                        None => println!("target cant resolve {}", ti),
                        Some(v) => {
                            domains.push(v.host);
                            port = v.port;
                        }
                    }
                }
                if domains.len() == 0 { continue }
                let dns_config = DNSConfig {
                    names: domains,
                    dns_type: "A".to_string(),
                    port,
                    refresh_interval: "10s".to_string(),
                };
                sc.dns_sd_configs.push(dns_config);
            }
            prometheus.scrape_configs.push(sc);
        }
        // 写入文件
        prometheus.flush();
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct Prometheus {
    global: Global,
    scrape_configs: Vec<ScrapeConfig>,
}

impl Prometheus {
    fn new() -> Prometheus {
        Prometheus {
            global: Global {
                scrape_interval: get_env("SCRAPE_INTERVAL".to_string(), "30s".to_string()),
                scrape_timeout: get_env("SCRAPE_TIMEOUT".to_string(), "10s".to_string()),
                evaluation_interval: get_env("EVALUATION_INTERVAL".to_string(), "10s".to_string()),
            },
            scrape_configs: Vec::new(),
        }
    }

    fn flush(&self) {
        let y = serde_yaml::to_string(self);
        let mut file = OpenOptions::new().read(true).write(true).create(true).open("prometheus.yml");
        file.unwrap().write_all(y.unwrap().as_bytes()).unwrap();
        println!("写入成功");
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct Global {
    scrape_interval: String,
    scrape_timeout: String,
    evaluation_interval: String,
}

#[derive(Debug, Serialize, Deserialize)]
struct ScrapeConfig {
    job_name: String,
    scrape_interval: String,
    scrape_timeout: String,
    metrics_path: String,
    static_configs: Vec<StaticConfig>,
    dns_sd_configs: Vec<DNSConfig>,
}

impl ScrapeConfig {
    fn new(name: String, metrics_path: String) -> ScrapeConfig {
        ScrapeConfig {
            job_name: name.to_string(),
            scrape_interval: "30s".to_string(),
            scrape_timeout: "10s".to_string(),
            metrics_path,
            static_configs: vec![],
            dns_sd_configs: vec![],
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct StaticConfig {
    targets: Vec<String>,
}

#[derive(Debug, Serialize, Deserialize)]
struct DNSConfig {
    names: Vec<String>,
    dns_type: String,
    port: u16,
    refresh_interval: String,
}

async fn load() {
    let mut jobs = Jobs::new();
    let interval = Duration::from_secs(10);
    loop {
        task::sleep(interval).await;
        // 加载配置 检测配置是否变化
        jobs.load()
    }
}

fn main() {
    task::block_on(load());
}
